﻿/**
 * estelime2.js
 *
 * estelionの機能をLimechat2のスクリプト用に書き直したもの。
 * 
 * ライセンス：このファイルのライセンスは、GPL(GNU Public License)とします。
 * 
 * @author  じーちゃん／へたっぴ
 * @license GPL http://www.gnu.org/licenses/gpl.html
 * @revision $Id: estelime2.js 2 2008-09-16 10:47:22Z  $
 */

var DEBUG = false;
var MAX_DICE_COUNT = 100;
var MAX_DICE_FACE  = 100000;
var GM_KEY = '0GM'; // GMデータを識別するためのキー。先頭に0につけている。
                    // これはIRCニックネームの先頭文字は数字を付けられないという
                    // レギュレーションを利用し、キー文字列が他人のニックネームと
                    // バッティングしないようにするため。

// 文頭にシャープ"#"がなくてもダイス反応するかどうか
// 1=反応する  0=反応しない
var ENABLE_SHARPLESS_DICE_FLAG = 1;

// BOT自身が入室した時点で、バージョン情報とヘルプURLを出力するかどうか
// 1=出力する  0=出力しない
var OUTPUT_ABOUT_AT_ENTER_ROOM_FLAG = 0;

var VERSION  = "1.2.2";

// リソース管理をする自動修得スキルの一覧
// インテンションはHPと絡むので、別処理
var autoLearnSkillHash = {
  "tb"  :"トルネード",
  "fami":"ファミリア",
  "dh"  :"ダンシング",
  "db"  :"ディビレテイト"
};
var autoLearnSkillRegExpStr;


// リソース管理をする自動修得スキル以外のスキル一覧
var skillHash = {
  "mc" :"メモリコンプリート",
  "hl" :"ハードラック",
  "sa" :"シークレット",
  "ad" :"アドヴァイス",
  "cm" :"カバームーブ",
  "pb" :"パーフェクト",
  "ub" :"アルティメット",
  "ac" :"アラウンドカバー",
  "sw" :"スピリチュアル",
  "pp" :"パリペイティア",
  "ps" :"パニッシャー",
  "dc" :"ダブルキャスト",
  "lb" :"リミットブレイク",
  "dg" :"デスゲイル",
  "ts" :"トップスピード",
  "sut":"スーパートリック",
  "stt":"ストップトラップ"
};
var skillRegExpStr;


var estelFuncsChannelArr;
var estelFuncsTalkArr;
var estelDataArr;
var gmNickToChannelArr;
var againstAttrArr;
var mt;

/*
 * 全角英数字を半角に変換する
 * Limechat Script: 文字コードはUTF-16LE
 */
var mb2sb = function(src) 
{
  var str = new String;
  var len = src.length;
  for (var i = 0; i < len; i++) {
    var c = src.charCodeAt(i);
    if (c >= 65281 && c <= 65374 && c != 65340) {
      str += String.fromCharCode(c - 65248);
    } else if (c == 0xf7) { /* ÷ */
      str += '/';
    } else if (c == 0xd7) { /* × */
      str += '*';
    } else if (c == 0x3000) { /* 　 */
      str += ' ';
    } else if (c == 0x3001) { /* 、 */
      str += ',';
    } else {
      str += src.charAt(i);
    } 
  }
  return str;
};


/*
 * スクリプトがロードされた時に実行される。
 * 初期化処理に使う。
 */
function event::onLoad() 
{
  // メルセンヌ・ツイスターの初期化
  mt = new MersenneTwister();
  
  // データ配列の初期化
  estelDataArr = [];
  gmNickToChannelArr = [];
  
  // 関数の登録
  estelFuncsChannelArr = new Array();
  estelFuncsChannelArr.push(estelCalcExp);
  estelFuncsChannelArr.push(estelInit);
  estelFuncsChannelArr.push(estelReset);
  estelFuncsChannelArr.push(estelKago);
  estelFuncsChannelArr.push(estelSet);
  estelFuncsChannelArr.push(estelShow);
  estelFuncsChannelArr.push(estelShowAll);
  estelFuncsChannelArr.push(estelWhereUsing);
  estelFuncsChannelArr.push(estelReport);
  estelFuncsChannelArr.push(estelIntension);
  estelFuncsChannelArr.push(estelDamage);
  estelFuncsChannelArr.push(estelGuild);
  estelFuncsChannelArr.push(estelGuildShow);
  estelFuncsChannelArr.push(estelGuildResetAll);
  estelFuncsChannelArr.push(estelSetGM);
  estelFuncsChannelArr.push(estelResetGM);
  estelFuncsChannelArr.push(estelRefreshAll);
  estelFuncsChannelArr.push(estelVersion);
  estelFuncsChannelArr.push(estelDice);
  estelFuncsChannelArr.push(estelGenerateRandomName);
  estelFuncsChannelArr.push(estelExport);
  estelFuncsChannelArr.push(estelKick);
  
  
  estelFuncsTalkArr = new Array();
  estelFuncsTalkArr.push(estelEnemySet);
  estelFuncsTalkArr.push(estelEnemyShow);
  estelFuncsTalkArr.push(estelEnemyShowAll);
  estelFuncsTalkArr.push(estelEnemyReset);
  estelFuncsTalkArr.push(estelEnemyDamage);
  
  
  // estelion管理のスキルをひっかけるための正規表現文字列の生成
  var tmpArr = Array();
  for (var key in autoLearnSkillHash) {
    tmpArr.push(key);
  }
  autoLearnSkillRegExpStr = tmpArr.join('|');
  
  tmpArr = Array();
  for (var key in skillHash) {
    tmpArr.push(key);
  }
  skillRegExpStr = tmpArr.join('|');
  
  // 対抗属性配列の初期化
  againstAttrArr = {
    "火" : "水",
    "水" : "火",
    "風" : "地",
    "地" : "風",
    "闇" : "光",
    "光" : "闇"
  };
  
  // 単体テスト
  // スクリプトが更新されたら自動的に走るようにここに登録
  if (DEBUG) { unitTest(); }
}


/*
 * チャンネル内でメッセージを送受信した時に実行されるイベント
 * estelime2の基本エントリ関数
 */
function event::onChannelText(prefix, channel, text)
{
  if (DEBUG) { log("prefix.nick:" + prefix.nick + "  prefix.username:" + prefix.username + "  channel:" + channel + "  text:" + text) }
  
  // estelionの仕様で、文頭が'#'で始まっていないものは、
  // コマンドとして認識させない
  // 
  // 2008/11/16追記：
  // ダイス振るのに文頭にシャープ付けることがないので、
  // ダイス振る機能だけシャープなしでコマンドうけつける設定を追加した。
  if (ENABLE_SHARPLESS_DICE_FLAG == 0) {
    if (!text.match(/^[#＃]/)) {
      return;
    }
  }
  
  // 基本的に各関数の中で、発言内容からコマンド判定をする
  // コマンド判定にマッチしなければfalseを返す。
  // コマンド判定にマッチしたら、処理をしてからtrueを返す
  var i;
  for (i = 0; i < estelFuncsChannelArr.length; i++) {
    if (estelFuncsChannelArr[i](prefix, channel, text)) {
      break;
    }
  }
}


/*
 * 招待されたとき実行されるイベント
 */
function event::onInvite(prefix, channel)
{
  join(channel);
  
  return true;
}


/*
 * チャンネルに入ったとき実行されるイベント
 */
function event::onJoin(prefix, channel)
{
  if (OUTPUT_ABOUT_AT_ENTER_ROOM_FLAG && prefix.nick == myNick) {
    estelSend(channel, "estelime2 Version:" + VERSION);
    estelSend(channel, "ドキュメントURL : http://www.g-chan.jp/trpg/wiki/?estelime2");
  }
  
  return true;
}


/*
 * 一対一のトークイベント
 */
function event::onTalkText(prefix, targetNick, text)
{
  if (DEBUG) { log("prefix.nick:" + prefix.nick + "  targetNick:" + targetNick + "  text:" + text) }
  
  // estelionの仕様で、文頭が'#'で始まっていないものは、
  // コマンドとして認識させない
  if (!text.match(/^#/)) {
    return;
  }
  
  // 基本的に各関数の中で、発言内容からコマンド判定をする
  // コマンド判定にマッチしなければfalseを返す。
  // コマンド判定にマッチしたら、処理をしてからtrueを返す
  var i;
  for (i = 0; i < estelFuncsTalkArr.length; i++) {
    if (estelFuncsTalkArr[i](prefix, text)) {
      break;
    }
  }
  
  return true;
}


/**
 * 発言する
 */
function estelSend(channelOrNick, mes)
{
  if (channelOrNick && mes.length > 0 
      && (channelOrNick.match(/^#/) || channelOrNick.length < 10)) {
    send(channelOrNick, mes);
  }
  if (DEBUG || !channelOrNick) {
    log(mes);
  }
}


/*
 * バージョン情報とヘルプのURLを出力する
 */
function estelVersion(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#(about|help)/)) {
    return false;
  }
  
  estelSend(channel, "<bold>estelime2 Version:<bold>" + VERSION);
  estelSend(channel, "<bold>ドキュメントURL :<bold> http://www.g-chan.jp/trpg/wiki/?estelime2");
  
  return true;
}


/*
 * ダイスのエラーチェック
 *
 * @return boolean true => OK false => NG
 */
function checkDiceError(channel, n, m)
{
  var errMes = null;
  
  if (n <= 0) {
    errMes = "0個以下のダイスを振ることはできません。";
  }
  if (n > MAX_DICE_COUNT) {
    errMes = MAX_DICE_COUNT + "個を超える数のダイスを振ることはできません。";
  }
  if (m < 2 || m > 100000) {
    errMes = "そのダイスを振ることができません。";
  }
  if (errMes) {
    estelSend(channel, errMes);
    return false;
  }
  return true;
}


/*
 * ダイスを振った計算結果の文字列を取得する
 *
 * javascriptだと、世間では「参照渡し」とか言ってるくせに、実は参照の値渡しなので、
 * 関数内で配列初期化しても関数抜けた後には反映されない。
 * なので、この関数に渡す retArr は事前に配列として初期化しておいてください。
 */
function getDiceResultStr(channel, rawDiceStr, retArr)
{
  if (!rawDiceStr.match(/(\d+)?(d|D|ar)(\d+)?/)) {
    return null;
  }
  
  var n;
  if (RegExp.$1.length > 0) {
    n = parseInt(RegExp.$1);
  } else {
    n = 1;
  }
  var m;
  if (RegExp.$2 == "ar") {
    m = 6;
  } else if (RegExp.$3.length > 0) {
    m = parseInt(RegExp.$3);
  } else {
    m = 6;
  }
  
  // エラー出力
  // 許容量を超えたダイスを振ろうとした場合には、エラーメッセージを出す。
  if (!checkDiceError(channel, n, m)) {
    return null;
  }
  
//  retArr = new Array();
  
  var i;
  var diceTotal = 0;
  var randRet;
  for (i = 0; i < n; i++) {
    randRet = mt.nextInt(0, parseInt(m)) + 1;
    diceTotal += randRet;
    retArr.push(randRet);
  }
  
  return diceTotal;
}


/*
 * クリティカルとなるダイス目の数を取得する
 */
var getCriticalDiceNum = function(diceArr)
{
  var i;
  var diceNum = 0;
  for (i = 0; i < diceArr.length; i++) {
    if (diceArr[i] == 6) {
      diceNum++;
    }
  }
  return diceNum;
};


/*
 * ファンブルとなるダイス目の数を取得する
 */
var getFunbleDiceNum = function(diceArr)
{
  var i;
  var diceNum = 0;
  for (i = 0; i < diceArr.length; i++) {
    if (diceArr[i] == 1) {
      diceNum++;
    }
  }
  return diceNum;
};


/*
 * ダイスを振った計算結果の文字列に変換する
 */
function convToDiceResultStr(channel, rawDiceStr)
{
  var diceStrNow = rawDiceStr;
  var retStr = rawDiceStr;
  var tmpStr;
  var diceArr;
  var diceVal;
  var tmpReplaceToL = "DICE";
  var tmpReplaceToS = "dice";
  var tmpReplaceToARA = "ARA";
  var totalARADiceNum = 0;
  var totalARACritDiceNum = 0;
  var totalARAFumbleDiceNum = 0;
  
  if (diceStrNow.match(/(d|D|ar)/i)) {
    while (diceStrNow.match(/((\d+)?((d|D|ar))(\d+)?)/)) {
      tmpStr = RegExp.$1;
      var diceNum = RegExp.$2;
      var araFlag = RegExp.$3 == "ar";
      var expandFlag = RegExp.$3 == "d";
      
      diceArr = new Array();
      diceVal = getDiceResultStr(channel, tmpStr, diceArr);
      if (diceVal == null || diceArr == null) {
        return false;
      }
      
      if (tmpStr.match(/d/)) {
        tmpStr2 = tmpStr.replace("d", tmpReplaceToS); // 一旦別の文字列に置換する
      }
      if (tmpStr.match(/D/)) {
        tmpStr2 = tmpStr.replace("D", tmpReplaceToL);
      }
      if (tmpStr.match(/ar/)) {
        tmpStr2 = tmpStr.replace("ar", tmpReplaceToARA);
      }
      if (araFlag) {
        totalARADiceNum += diceNum;
        totalARACritDiceNum += getCriticalDiceNum(diceArr);
        totalARAFumbleDiceNum += getFunbleDiceNum(diceArr);
      }
      
      retReplaceStr = diceVal;
      if (expandFlag || araFlag) {
        retReplaceStr += "(" + tmpStr2 + ": " + diceArr.join(" ") + ")";
      } else {
        retReplaceStr += "(" + tmpStr2 + ")";
      }
      retStr = retStr.replace(tmpStr, retReplaceStr);
      diceStrNow = diceStrNow.replace(tmpStr, diceVal);
    }
  }
  
  regExp = new RegExp(tmpReplaceToS);
  while (retStr.match(regExp)) {
    retStr = retStr.replace(tmpReplaceToS, "d");
  }
  regExp = new RegExp(tmpReplaceToL);
  while (retStr.match(regExp)) {
    retStr = retStr.replace(tmpReplaceToL, "D");
  }
  regExp = new RegExp(tmpReplaceToARA);
  while (retStr.match(regExp)) {
    retStr = retStr.replace(tmpReplaceToARA, "ar");
  }
  
  retStr = retStr + " = <bold>" + eval(diceStrNow) + "<bold>";
  
  if (totalARADiceNum >= 2) {
    if (totalARACritDiceNum >= 2) {
      retStr += "  <color blue>クリティカル!! 6の数:" + totalARACritDiceNum + "個<stop>";
    } else if (totalARADiceNum == totalARAFumbleDiceNum) {
      retStr += "  <color red>ファンブル!!<stop>";
    }
  }
  
  return retStr;
}


/* 
 * ダイスを振る
 */
function estelDice(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  
  // 全角を半角に変換
  text = mb2sb(text);
  
  if (text.match(/^#?([\d\.\(\)]+)($|\s+([^\s]*)$)/)) {
    // ダイスも計算式もなければなんもしなくていいっしょ？
    return false;
  }
  if (!text.match(/^#?(([\d\.dD\(\)\+\-\*\/]|ar)+)($|\s+([^\s]*)$)/)) {
    return false;
  }
  
  var diceStr = RegExp.$1;
  var mesStr  = RegExp.$3;
  
  var convedDiceStr = convToDiceResultStr(channel, diceStr);
  var sendStr = '';
  if (prefix && prefix.nick != "undefined") {
    sendStr += prefix.nick + " : ";
  }
  sendStr += diceStr + " => " + convedDiceStr + "  " + mesStr;
  
  estelSend(channel, sendStr);
  
  return true;
}


/*
 * 成長に必要な経験値を計算する
 */
function estelCalcExp(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#calcexp\s+((\s+|(lv|cl)=(\d+)|change=(\d+)|fate=(\d+))+)/i)) {
    return false;
  }
  
  var lv = 0;
  var change = 0;
  var fate = 0;
  
  if (RegExp.$4) lv = RegExp.$4;
  if (RegExp.$5) change = RegExp.$5;
  if (RegExp.$6) fate = RegExp.$6;
  
  var lvexp = 0;
  var i;
  for (i = lv - 1; i > 0; i--) {
    lvexp += i * 10;
  }
  var fateexp = 0;
  for (i = fate - 5; i > 0; i--) {
    fateexp += (i + 4) * 2;
  }
  var changeexp = change * 10;
  var total = changeexp + fateexp + lvexp;
  
  var sendStr = "Lv:" + lv + " (used:" + lvexp + ") FATE:" + fate + 
                " (used:" + fateexp + ") JobChange:" + change + 
                " (used:" + changeexp + ") TOTAL:" + total;
  if (RegExp.$7) sendStr += ' ' + RegExp.$7;
  
  estelSend(channel, sendStr);
  
  return true;
}


/*
 * 数値型のデータかどうかを自動的に判別してセットする
 */
function setDataAsInt(arr, key, val)
{
  if (isNaN(val)) {
    arr[key] = val;
  } else {
    arr[key] = parseInt(val);
  }
}


/*
 * estelDataArrにデータをセットする
 *
 * Perlと違ってJavaScriptは、利用前に配列初期化してやらないとエラーになるので
 * データセットするラッパー関数を用意
 */
function setEstelDataArr(channel, nick, key, val)
{
  if (!estelDataArr[channel]) {
    estelDataArr[channel] = [];
  }
  if (nick) {
    if (!estelDataArr[channel][nick]) {
      estelDataArr[channel][nick] = [];
    }
    setDataAsInt(estelDataArr[channel][nick], key ,val);
    if (estelDataArr[channel][nick]['nick'] == undefined) {
      estelDataArr[channel][nick]['nick'] = nick;
    }
  } else {
    setDataAsInt(estelDataArr[channel], key ,val);
  }
}


/*
 * 実際の初期化処理関数
 */
function doInit(prefix, channel)
{
  // GMニックネームからチャンネルを引くための配列からデータ削除
  deleteGMNickToChannel(channel);
  
  delete estelDataArr[channel];
  
  var sendStr = '"' + channel + '" のデータを初期化しました。';
  estelSend(channel, sendStr);
}


/*
 * 機能全体初期化処理のエントリ関数
 */
function estelInit(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#(resetall|init)\s*$/i)) {
    return false;
  }
  
  var time = Math.floor(new Date().getTime() / 1000); // ミリ秒単位なので秒に変換
  if (estelDataArr[channel] != undefined
      && estelDataArr[channel]['previnittime'] > 0 
      && time - estelDataArr[channel]['previnittime'] < 3*60) // 3*60=3分
  {
    // 初期化実行
    doInit(prefix, channel);
  } else {
    setEstelDataArr(channel, null, 'previnittime', time);
    var sendStr = '本当に "' + channel + '" のデータを初期化しますか？' +
                  ' 初期化する場合には、3分以内にもう一度 #init を実行して下さい。';
    estelSend(channel, sendStr);
  }
  
  return true;
}


/*
 * 行動値の数値部分を固定文字列に調整
 */
function adjustNick(nick)
{
  if (nick.match(/^([^\d]\d+_)/)) {
    var tmpStr = RegExp.$1;
    nick = nick.replace(tmpStr, "");
  }
  return nick;
}


/*
 * 一人分のデータリセット
 */
function estelReset(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#reset\s*$/i)) {
    return false;
  }
  if (!prefix || !prefix.nick) {
    return false;
  }
  if (estelDataArr[channel] == undefined) {
    return false;
  }
  
  var nick = adjustNick(prefix.nick);
  delete estelDataArr[channel][nick];
  
  var sendStr = '"' + prefix.nick + '" のデータを初期化しました。';
  estelSend(channel, sendStr);

  return true;
}


/*
 * ギルドスキル加護の設定
 * （ギルドスキルがいろいろ増えた今になって、あえて加護って使われるのか？ｗ）
 */
function estelKago(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#kago(\s+|=)((.+)\/)?(\d+)$/i)) {
    return false;
  }
  
  var guildName = RegExp.$3;
  var kago = parseInt(RegExp.$4);
  var sendStr = '';
  
  if (guildName) {
    // ギルドあり
    createGuildData(channel, guildName);
    if (estelDataArr[channel]['guild'][guildName]['skillStr'] == undefined) {
      estelDataArr[channel]['guild'][guildName]['skillStr'] =
        '加護' + kago;
    }
    estelDataArr[channel]['guild'][guildName]['kago'] = kago;
    
    sendStr = 'ギルド "' + guildName + '" に加護' + kago + 'を設定しました。';
    estelSend(channel, sendStr);
  } else {
    // ギルドなし
    if (estelDataArr[channel] != undefined
        && estelDataArr[channel]['guild'] != undefined) {
      sendStr = 'ギルドデータが登録されています。' +
                'ギルドデータが登録されている場合には' +
                '全体に対する加護設定は無効となります。';
      estelSend(channel, sendStr);
      return false;
    }
    setEstelDataArr(channel, null, 'kago', kago);
    
    sendStr = '加護' + estelDataArr[channel]['kago'] + 'を設定しました。';
    estelSend(channel, sendStr);
  }
  
  return true;
}


/*
 * 各データの出力
 */
function outputData(channelOrNick, arr, isVerbose)
{
  var sendStr = '';
  
  // reset直後とか、データがまったく設定されていない場合には
  // 何もしないで帰る
  if (arr == undefined) {
    return;
  }
  
  if (arr['id'] != undefined) {
    sendStr += 'ID:' + arr['id'] + ' ';
  }
  if (arr['name'] != undefined) {
    sendStr += '名前:' + arr['name'];
    if (arr['nick'] != undefined) {
      var dispNick = arr['nick'].replace('_XX_', '');
      sendStr += '(' + dispNick + ')';
    }
  }
  if (arr['attr'] != undefined) {
    sendStr += ' 属性:' + arr['attr'];
  }
  if (arr['pl'] != undefined) {
    sendStr += ' PL:' + arr['pl'];
  }
  if (arr['cl'] != undefined) {
    sendStr += ' CL:' + arr['cl'];
  }
  if (arr['fate'] != undefined) {
    sendStr += ' FATE:' + arr['fate'] + '/' + arr['maxfate'];
    if (arr['fatesk'] != undefined) {
      sendStr += '(+' + arr['fatesk'] + ')';
    }
  }
  if (arr['maxhp'] != undefined
      && arr['hp'] != undefined) {
    sendStr += ' HP:' + arr['hp'] + '/' + arr['maxhp'];
  }
  if (arr['maxmp'] != undefined
      && arr['mp'] != undefined) {
    sendStr += ' MP:' + arr['mp'] + '/' + arr['maxmp'];
    if (arr['fami'] != undefined) {
      sendStr += '+(' + arr['fami'] + ')';
    }
  }
  if (arr['def'] != undefined) {
    sendStr += ' 物防:' + arr['def'];
  }
  if (arr['mdef'] != undefined) {
    sendStr += ' 魔防:' + arr['mdef'];
  }
  
  var key;
  
  // 自動修得スキル系
  for (key in autoLearnSkillHash) {
    if ('fami' == key) {
      // ファミリアの場合はここでは表示させない
      continue;
    }
    if (arr[key] != undefined) {
      if (isVerbose) {
        sendStr += ' ' + autoLearnSkillHash[key] + ':' + arr[key];
      } else {
        sendStr += ' ' + key + ':' + arr[key];
      }
    }
  }
  
  // その他スキル系
  for (key in skillHash) {
    if (arr[key] != undefined) {
      if (isVerbose) {
        sendStr += ' ' + skillHash[key] + ':' + arr[key];
      } else {
        sendStr += ' ' + key + ':' + arr[key];
      }
    }
  }
  
  estelSend(channelOrNick, sendStr);
}


/*
 * 各データのexport用出力
 */
function exportData(channelOrNick, arr)
{
  var sendStr = '#set ';
  
  // reset直後とか、データがまったく設定されていない場合には
  // 何もしないで帰る
  if (arr == undefined) {
    return;
  }
  
  if (arr['name'] != undefined) {
    sendStr += ' name=' + arr['name'];
  }
  if (arr['pl'] != undefined) {
    sendStr += ' pl=' + arr['pl'];
  }
  if (arr['cl'] != undefined) {
    sendStr += ' cl=' + arr['cl'];
  }
  if (arr['maxfate'] != undefined) {
    sendStr += ' maxfate=' + arr['maxfate'];
  }
  if (arr['fate'] != undefined) {
    sendStr += ' fate=' + arr['fate'];
  }
  if (arr['fatesk'] != undefined) {
    sendStr += ' fatesk=' + arr['fatesk'];
  }
  if (arr['maxhp'] != undefined
      && arr['hp'] != undefined) {
    sendStr += ' hp=' + arr['hp'] + ' maxhp=' + arr['maxhp'];
  }
  if (arr['maxmp'] != undefined
      && arr['mp'] != undefined) {
    sendStr += ' mp=' + arr['mp'] + ' maxmp=' + arr['maxmp'];
  }
  if (arr['def'] != undefined) {
    sendStr += ' def=' + arr['def'];
  }
  if (arr['mdef'] != undefined) {
    sendStr += ' mdef=' + arr['mdef'];
  }
  
  var key;
  
  // 自動修得スキル系
  for (key in autoLearnSkillHash) {
    if (arr[key] != undefined) {
      sendStr += ' ' + key + '=' + arr[key];
    }
  }
  
  // その他スキル系
  for (key in skillHash) {
    if (arr[key] != undefined) {
      sendStr += ' ' + key + '=' + arr[key];
    }
  }
  
  estelSend(channelOrNick, sendStr);
}


/*
 * 数値の加減算部分
 */
function calcEstelData(arr, key, val, vec, channelOrNick)
{
  if (arr == undefined
      || arr[key] == undefined) {
    // データが登録されていないよエラー
    sendStr = 'エラー。データが登録されていません。';
    estelSend(channelOrNick, sendStr);
    
    return false;
  }
  
  var tmpVal = parseInt(arr[key]);
  if (isNaN(tmpVal)) {
    // 数値変換できない項目に対して加減算されたらエラー
    sendStr = 'エラー。その項目は加減算を行えません。';
    estelSend(channelOrNick, sendStr);
    
    return false;
  }
  
  // フェイト関連チェック
  var skipFlag = false;
  if (key == 'fate') {
    var tmpFate = arr['fate'];
    if (arr['fatesk'] != undefined) {
      tmpFate += arr['fatesk'];
    }
    tmpFate += val * vec;
    
    // フェイト残量check
    if (tmpFate < 0) {
      // フェイト残量不足
      sendStr = 'エラー。フェイトが足りません。';
      estelSend(channelOrNick, sendStr);
      
      return false;
    }
    
    // まず、普通のフェイトから先に使用分を減らす
    arr['fate'] += val * vec;
    
    // そして、普通のフェイトの残量が不足したら自動的にスキルのフェイトから減らす
    if (arr['fate'] < 0) {
      arr['fatesk'] += arr['fate'];
      arr['fate'] = 0;
    }
    skipFlag = true;
  }
  
  if (!skipFlag) {
    tmpVal += val * vec;
    arr[key] = tmpVal;
  }
  
  // maxhp, maxmp check
  if (key == 'hp' && arr['hp'] > arr['maxhp']) {
    arr['hp'] = arr['maxhp'];
  }
  if (key == 'mp' && arr['mp'] > arr['maxmp']) {
    arr['mp'] = arr['maxmp'];
  }
  
  return true;
}


/*
 * 加減算の種類の取得
 */
function getVec(text)
{
  if (text.match(/\+=/)) {
    return 1;
  } else if (text.match(/-=/)) {
    return -1;
  }
  return 0;
}


/*
 * 各種数値データの設定
 */
function estelSet(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  
  // 正直、やり過ぎ感漂う正規表現……
  var regExp = new RegExp(
    "^#(set\\s+)?((" + 
      "((name|pl|guild)=[^\\s]+)|" + 
      "(cl|lv|maxfate|fate|maxhp|hp|mhp|maxmp|mp|mmp|def|mdef|fatesk)[+-]?=\\d+|" +
      "(" + autoLearnSkillRegExpStr + ")[+-]?=(\\d+|on)|" +
      "(" + skillRegExpStr + ")[+-]?=\\d+|" +
      "\\s" + 
    ")+)$", "i"
  );
  
  if (!text.match(regExp)) {
    return false;
  }
  
  var setFlag = RegExp.$1 != '';
  var setArr  = RegExp.$2.toLowerCase().split(/\s+/);
  var sendStr;
  var nick = adjustNick(prefix.nick);
  
  for (var i in setArr) {
    // 値の設定(=)か、加減算(+=/-=)かを取得
    var vec = getVec(setArr[i]);
    
    // キーと値の取得
    var tmpArr = setArr[i].split(/[\+-]?=/);
    var key = tmpArr[0];
    var val = tmpArr[1];
    
    // "lv"は"cl"に直す
    if (key == "lv") key = "cl";
    
    // "mhp"は"maxhp"に直す
    if (key == "mhp") key = "maxhp";
    
    // 最大HP"maxhp"の場合で、現在HPが設定されていない場合は
    // これも一緒に設定しておく（最初の一回のみ）
    if (key == "maxhp" && estelDataArr[channel][nick]['hp'] == undefined) {
      setEstelDataArr(channel, nick, 'hp', val);
    }
    
    // "mmp"は"maxmp"に直す
    if (key == "mmp") key = "maxmp";
    
    // 最大MP"maxmp"の場合で、現在MPが設定されていない場合は
    // これも一緒に設定しておく（最初の一回のみ）
    if (key == "maxmp" && estelDataArr[channel][nick]['mp'] == undefined) {
      setEstelDataArr(channel, nick, 'mp', val);
    }
    
    // フェイト"fate"の場合で、表記用のフェイト最大値が設定されていない場合には
    // これも一緒に設定しておく（最初の一回のみ）
    if (key == "fate" && estelDataArr[channel][nick]['maxfate'] == undefined) {
      setEstelDataArr(channel, nick, 'maxfate', val);
    }
    
    
    // 自動修得スキル関連
    for (var autoLearnSkillKey in autoLearnSkillHash) {
      if (key == autoLearnSkillKey) {
        if (val == "on") {
          // "on"によってclから自動的に値をセットする場合には
          // 事前にclが設定されてるかチェック
          if (!(estelDataArr[channel][nick]['cl'] != undefined
              && estelDataArr[channel][nick]['cl'] > 0)) {
            // clが設定されていない
            sendStr = 'エラー。clが正しく設定されていません。';
            estelSend(channel, sendStr);
            
            return false;
          }
        }
        
        if (!setFlag && val == "on") {
          // “#set”コマンドを経由しない場合、解放コマンドとして取り扱うので
          // ちゃんと事前にfateが登録されているかどうかをチェック
          if (estelDataArr[channel][nick]['fate'] == undefined) {
            // fateが設定されていない
            sendStr = 'エラー。フェイトが設定されていません。';
            estelSend(channel, sendStr);
            
            return false;
          } else if (estelDataArr[channel][nick]['fate'] < 1) {
            // 通常のフェイトが足りない場合は、スキルフェイトをチェックする
            if (estelDataArr[channel][nick]['fatesk'] != undefined
                && estelDataArr[channel][nick]['fatesk'] > 0) {
              estelDataArr[channel][nick]['fatesk']--;
            } else {
              // fateが足りない
              sendStr = 'エラー。フェイトが足りません。';
              estelSend(channel, sendStr);
              
              return false;
            }
          } else {
            // ちゃんとフェイトが1以上ある状態なら、それを減らす。
            estelDataArr[channel][nick]['fate']--;
          }
          sendStr = autoLearnSkillHash[autoLearnSkillKey] +
                   'を解放しました。フェイトを1減らします。';
          estelSend(channel, sendStr);
        }
        
        if (val == "on") {
          var cl = estelDataArr[channel][nick]['cl'] > 10 ? 
                   10 : estelDataArr[channel][nick]['cl'];
          if (key == "fami") {
            // ファミリアだけは別扱い
            val = cl * 10;
          } else {
            val = cl;
          }
        }
      }
    }
    
    if (vec != 0) {
      // 加減算の実行
      if (!calcEstelData(estelDataArr[channel][nick], key, val, vec, channel)) {
        return false;
      }
    } else {
      // 値のセット
      setEstelDataArr(channel, nick, key, val);
      
      // 値のセットが終わったあとに、所属ギルドの情報が更新されていたら
      // ギルドの結びつけも更新する
      if (key == 'guild') {
        connectCharToGuild(channel, nick);
      }
    }
  }
  
  // 設定したデータを出力
  outputData(channel, estelDataArr[channel][nick], false);
  
  return true;
}


/*
 * データ出力
 */
function estelShow(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#show(\s+v)?\s*$/i)) {
    return false;
  }
  var isVerbose = RegExp.$1 ? true : false;
  
  // 設定したデータを出力
  var nick = adjustNick(prefix.nick);
  outputData(channel, estelDataArr[channel][nick], isVerbose);
  
  return true;
}


/*
 * データ出力（全員分）
 */
function estelShowAll(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#showall(\s+v)?\s*$/i)) {
    return false;
  }
  if (estelDataArr[channel] == undefined) {
    return false;
  }
  var isVerbose = RegExp.$1 ? true : false;
  
  // 設定したデータを出力
  for (var nick in estelDataArr[channel]) {
    outputData(channel, estelDataArr[channel][nick], isVerbose);
  }
  
  return true;
}


/*
 * データエクスポート（全員分）
 */
function estelExport(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#export\s*$/i)) {
    return false;
  }
  if (estelDataArr[channel] == undefined) {
    return false;
  }
  
  // 設定したデータを出力
  for (var nick in estelDataArr[channel]) {
    exportData(channel, estelDataArr[channel][nick]);
  }
  
  return true;
}


/*
 * どこのチャンネルで使用されているかを表示する
 * ……クライアントBOTなら、もうこの機能いらないんじゃないかなぁ？
 */
function estelWhereUsing(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#whereusing\s*$/i)) {
    return false;
  }
  
  // 設定したデータを出力
  var sendStr = '';
  var tmpArr = new Array();
  for (var usedChannel in estelDataArr) {
    tmpArr.push(usedChannel);
  }
  
  if (tmpArr.length > 0) {
    sendStr += '現在 ' + tmpArr.join(', ') + ' にて使用されています。';
  } else {
    sendStr += '現在使用されていません。';
  }
  estelSend(channel, sendStr);
  
  return true;
}


/*
 * セッション終了後の成長点レポート生成
 */
function estelReport(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#report(ng)?\s+(\d+)(\s+all)?/i)) {
    return false;
  }
  
  var ng  = RegExp.$1 != '';
  var all = RegExp.$3 != '';
  var exp = parseInt(RegExp.$2);
  var sendStr = '';
  var totalGuildExp = 0;
  var totalGMExpArr = new Array();
  var playerCount = 0;
  var guildExpArr = new Array();
  
  for (var nick in estelDataArr[channel]) {
    // チャンネルに結びつくデータは、キャラクターのデータ以外にもいくつかある。
    // その中から、キャラクターデータならfateは設定されているはず、と
    // 決め打ちで判別する。maxfateが設定されていない場合はキャラデータとみなさない。
    if (estelDataArr[channel][nick]['maxfate'] == undefined) {
      continue;
    }
    
    // ギルドデータが設定されている場合
    var guildName = '';
    if (estelDataArr[channel]['guild']) {
      if (estelDataArr[channel][nick]['guild'] == undefined) {
        // キャラの所属ギルドが設定されていない場合にはエラーとする
        sendStr = 'エラー。' + estelDataArr[channel][nick]['name'] + 
                  'の所属ギルドが登録されていません。';
        estelSend(channel, sendStr);
        return false;
      }
      guildName = estelDataArr[channel][nick]['guild'];
      if (estelDataArr[channel]['guild'][guildName] == undefined) {
        // キャラの所属ギルドがギルド登録されていない
        sendStr = 'エラー。' + estelDataArr[channel][nick]['name'] + 
                  'の所属ギルド "' + guildName + '" がギルドデータとして' +
                  '登録されていません。';
        estelSend(channel, sendStr);
        return false;
      }
    }
    
    var total = 0;
    var maxfate = estelDataArr[channel][nick]['maxfate'];
    var usedFate = 0;
    if (all) {
      usedFate = maxfate;
    } else {
      usedFate = maxfate - estelDataArr[channel][nick]['fate'];
    }
    
    if (ng) {
      // ギルド上納なし
      total = exp + usedFate;
      sendStr = estelDataArr[channel][nick]['name'];
      if (estelDataArr[channel][nick]['pl'] != undefined) {
        sendStr += '(' + estelDataArr[channel][nick]['pl'] + ')';
      }
      sendStr += ' CL:' + estelDataArr[channel][nick]['cl'] +
                 ' FATE:' + usedFate +
                 ' EXP:' + (exp + usedFate) +
                 ' TOTAL:' + total;
    } else {
      // ギルド上納あり
      total = exp + usedFate - estelDataArr[channel][nick]['cl'];
      if (guildName) {
        if (guildExpArr[guildName] == undefined) {
          guildExpArr[guildName] = estelDataArr[channel][nick]['cl']
        } else {
          guildExpArr[guildName] += estelDataArr[channel][nick]['cl']
        }
      } else {
        totalGuildExp += estelDataArr[channel][nick]['cl'];
      }
      
      sendStr = estelDataArr[channel][nick]['name'];
      if (estelDataArr[channel][nick]['pl'] != undefined) {
        sendStr += '(' + estelDataArr[channel][nick]['pl'] + ')';
      }
      sendStr += ' CL:' + estelDataArr[channel][nick]['cl'] +
                 ' FATE:' + usedFate +
                 ' EXP:' + (exp + usedFate) +
                 ' PAY:' + estelDataArr[channel][nick]['cl'] +
                 ' TOTAL:' + total;
    }
    estelSend(channel, sendStr);
    
    totalGMExpArr.push(total);
    playerCount++;
  }
  
  // GM成長点
  var bunbo = playerCount > 3 ? 3 : playerCount;
  sendStr = 'GM EXP:(' + totalGMExpArr.join('+') + ')/' + bunbo + '+1=';
  sendStr += (Math.floor(eval(totalGMExpArr.join('+')) / bunbo) + 1);
  estelSend(channel, sendStr);
  
  if (!ng) {
    // ギルド上納成長点
    sendStr = '';
    if (estelDataArr[channel]['guild']) {
      for (var guildName in guildExpArr) {
        sendStr = guildName + " EXP:" + guildExpArr[guildName];
        estelSend(channel, sendStr);
      }
    } else {
      sendStr = "Guild EXP:" + totalGuildExp;
      estelSend(channel, sendStr);
    }
  }
  
  return true;
}


/*
 * モンクの自動修得スキル「インテンション」の処理
 */
function estelIntension(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#(pre)?int(=(on|\d+))?/i)) {
    return false;
  }
  
  var nick = adjustNick(prefix.nick);
  var preFlag = RegExp.$1;
  
  // "on"によってclから自動的に値をセットする場合には
  // 事前にclが設定されてるかチェック
  if (!(estelDataArr[channel][nick]['cl'] != undefined
      && estelDataArr[channel][nick]['cl'] > 0)) {
    // clが設定されていない
    sendStr = 'エラー。clが正しく設定されていません。';
    estelSend(channel, sendStr);
    
    return false;
  }
  
  // ちゃんと事前にfateが登録されているかどうかをチェック
  if (estelDataArr[channel][nick]['fate'] == undefined) {
    // fateが設定されていない
    sendStr = 'エラー。フェイトが設定されていません。';
    estelSend(channel, sendStr);
    
    return false;
  } else if (estelDataArr[channel][nick]['fate'] < 1) {
    // 通常のフェイトが足りない場合は、スキルフェイトをチェックする
    if (estelDataArr[channel][nick]['fatesk'] != undefined
        && estelDataArr[channel][nick]['fatesk'] > 0) {
      estelDataArr[channel][nick]['fatesk']--;
    } else {
      // fateが足りない
      sendStr = 'エラー。フェイトが足りません。';
      estelSend(channel, sendStr);
      
      return false;
    }
  } else {
    // ちゃんとフェイトが1以上ある状態なら、それを減らす。
    estelDataArr[channel][nick]['fate']--;
  }
  
  cl = 0;
  if (text.match(/=(\d+)/)) {
    cl = RegExp.$1;
  } else {
    cl = estelDataArr[channel][nick]['cl'] > 10 ? 
         10 : estelDataArr[channel][nick]['cl'];
  }
  estelDataArr[channel][nick]['maxhp'] += 5 * cl;
  
  var sendStr = '';
  
  if (preFlag) {
    estelDataArr[channel][nick]['hp'] += 5 * cl;
    sendStr = 'インテンションを解放しました。HP(現在値)とHP最大値が' +
              (5 * cl) + '増加しました。';
    estelSend(channel, sendStr);
  } else {
    sendStr = 'インテンションを解放しました。HP最大値が' +
              (5 * cl) + '増加しました。';
    estelSend(channel, sendStr);
  }
  
  sendStr = 'フェイトを1減らしました。';
  estelSend(channel, sendStr);
  
  outputData(channel, estelDataArr[channel][nick], false);
  
  return true;
}


/*
 * 防御点の取得
 * エラー発生したら -1 返します。0以上返ってきたらそれが防御点。
 */
function getDef(arr, mdamFlag)
{
  var def = 0;
  var sendStr = '';
  
  if (!mdamFlag) {
    if (arr['def'] == undefined) {
      // 物理防御が設定されてなければエラー
      sendStr = 'エラー。物理防御の値が設定されていません。';
      estelSend(channel, sendStr);
      
      return -1;
    }
    def = arr['def'];
  } else {
    if (arr['mdef'] == undefined) {
      // 魔法防御が設定されてなければエラー
      sendStr = 'エラー。魔法防御の値が設定されていません。';
      estelSend(channel, sendStr);
      
      return -1;
    }
    def = arr['mdef'];
  }
  return def;
}


/*
 * ダメージ適用後の出力文字列の取得
 * 
 * 引数の数が多すぎてダサいけど、このままいきまする
 */
function getDamageSendStr(arr, def, mdamFlag, damage, prevent, kago, kagoGuildName,
                          hpDamage, attkAttr, charAttr)
{
  var sendStr;
  
  if (!mdamFlag) {
    sendStr = '物理ダメージ:' + damage;
  } else {
    sendStr = '魔法ダメージ';
    if (attkAttr) {
      sendStr += '(' + attkAttr + ')';
    }
    sendStr += ':' + damage;
  }
  if (prevent) {
    sendStr += ' 軽減:' + prevent;
  }
  if (kago) {
    sendStr += ' 加護';
    if (kagoGuildName) {
      sendStr += '[' + kagoGuildName + ']';
    }
    sendStr += '(' + kago + '):' + (kago * 2);
  }
  if (!mdamFlag) {
    sendStr += ' 物防:' + def;
  } else {
    sendStr += ' 魔防:' + def;
    if (attkAttr && attkAttr != '－') {
      if (attkAttr == charAttr) {
        sendStr += '(同属性により2倍)';
      } else if (attkAttr == againstAttrArr[charAttr]) {
        sendStr += '(対抗属性により0)';
      }
    }
  }
  sendStr += ' 最終ダメージ:' + hpDamage;
  sendStr += ' ' + arr['name'] + 'の残りHP:' + arr['hp'];
  
  return sendStr;
}


/*
 * ダメージの適用
 */
function estelDamage(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#(m?)dam\s((\s|(\d+)|\sp=?(\d+)|\skago=?(\d+))+)$/i)) {
    return false;
  }
  
  var mdamFlag = RegExp.$1 != '';
  var damage  = parseInt(RegExp.$4);
  var prevent = parseInt(RegExp.$5);
  var kago    = parseInt(RegExp.$6);
  var nick = adjustNick(prefix.nick);
  var sendStr = '';
  
  if (damage == '') {
    // ダメージの値がおかしい
    sendStr = 'エラー。ダメージの値が正しく設定されていません。';
    estelSend(channel, sendStr);
    
    return false;
  }
  
  // 現在HPが登録されているかどうかチェック
  if (estelDataArr[channel] == undefined
      || estelDataArr[channel][nick] == undefined
      || estelDataArr[channel][nick]['hp'] == undefined) {
    // 現在HPが登録されてないよエラー
    sendStr = 'エラー。現在HPの値が設定されていません。';
    estelSend(channel, sendStr);
    
    return false;
  }
  
  // 入力された式に加護が設定されていなければ、事前登録されている加護を適用
  if (isNaN(kago)
      && estelDataArr[channel]['kago'] != undefined
      && !isNaN(estelDataArr[channel]['kago'])) {
    kago = estelDataArr[channel]['kago'];
  }
  
  var kagoGuildName = null;
  if (isNaN(kago)) {
    if (estelDataArr[channel]['guild'] != undefined) {
      // ギルドデータが1件でも登録されていれば、ギルドのほうの加護を使う
      if (estelDataArr[channel][nick]['guild'] != undefined) {
        var guildName = estelDataArr[channel][nick]['guild'];
        if (estelDataArr[channel]['guild'][guildName] != undefined
            && estelDataArr[channel]['guild'][guildName]['kago'] != undefined) {
          // ふぅ。ここまで判定が通ってようやくギルドの加護が使えるぜぃ
          kago = estelDataArr[channel]['guild'][guildName]['kago'];
          kagoGuildName = guildName;
        }
      }
    } else {
      // ギルドデータなし
      if (estelDataArr[channel]['kago'] != undefined
          && !isNaN(estelDataArr[channel]['kago'])) {
        kago = estelDataArr[channel]['kago'];
      }
    }
  }
  
  var def = getDef(estelDataArr[channel][nick], mdamFlag);
  if (def < 0) {
    // 防御の値がおかしい
    var tmpStr = '物理';
    if (mdefFlag) {
      tmpStr = '魔法';
    }
    sendStr = 'エラー。' + tmpStr + '防御の値が正しく設定されていません。';
    estelSend(channel, sendStr);
    return false;
  }
  
  // HPダメージ値の決定
  if (isNaN(prevent)) prevent = 0;
  if (isNaN(kago)) kago = 0;
  var hpDamage = damage - prevent - kago * 2 - def;
  if (hpDamage < 0) {
    hpDamage = 0;
  }
  estelDataArr[channel][nick]['hp'] -= hpDamage;
  
  sendStr = getDamageSendStr(estelDataArr[channel][nick], def, mdamFlag, 
                             damage, prevent, kago, kagoGuildName, hpDamage,
                             null, null);
  
  estelSend(channel, sendStr);
  
  return true;
}


/*
 * 全員HPかMP（または両方）を全快させる
 */
function estelRefreshAll(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#refreshall\s+(hp|mp|both)$/i)) {
    return false;
  }
  var refreshTarget = RegExp.$1;
  
  if (estelDataArr[channel] == undefined) {
    return false;
  }
  
  var sendStr = '';
  sendStr = '回復前（元に戻す必要がある場合には以下のデータから手動でお願いします）';
  estelSend(channel, sendStr);
  
  estelShowAll(prefix, channel, "#showall");
  
  // 設定したデータを出力
  for (var nick in estelDataArr[channel]) {
    var charData = estelDataArr[channel][nick];
    switch (refreshTarget) {
    case 'hp':
      charData['hp'] = charData['maxhp'];
      break;
    
    case 'mp':
      charData['mp'] = charData['maxmp'];
      break;
    
    case 'both':
      charData['hp'] = charData['maxhp'];
      charData['mp'] = charData['maxmp'];
      break;
    }
  }
  
  sendStr = '　';
  estelSend(channel, sendStr);
  sendStr = '回復後';
  estelSend(channel, sendStr);
  
  estelShowAll(prefix, channel, "#showall");
  
  return true;
}


/*
 * 退出させる
 */
function estelKick(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#kick\s*$/i)) {
    return false;
  }
  
  part(channel, 'ではまた～　ノシ');
  
  return true;
}


/*
 * キャラクターをギルドに結びつける
 */
function connectCharToGuild(channel, nick)
{
  if (estelDataArr[channel] == undefined
      || estelDataArr[channel]['guild'] == undefined) {
    // guildが登録されていなければすることないんで帰る
    return false;
  }
  if (estelDataArr[channel][nick]['guild'] == undefined
      || estelDataArr[channel][nick]['name'] == undefined) {
    // guildと名前が登録されていないキャラデータには用はない
    return false;
  }
  
  // 登録されているギルドでループを回す
  // キャラに設定されているギルド名が、登録されているギルド名と一致して
  // かつ、まだギルドにキャラ名が登録されていなければ登録する。
  for (var guildName in estelDataArr[channel]['guild']) {
    var guild = estelDataArr[channel]['guild'][guildName];
    var skipFlag = false;
    
    if (estelDataArr[channel][nick]['guild'] == guildName) {
      // キャラに登録されているギルド
      // すでに登録済みでなければ、追加登録する
      for (var i in guild['charArr']) {
        if (guild['charArr'][i] == estelDataArr[channel][nick]['name']) {
          // すでに登録済み
          skipFlag = true;
          break;
        }
      }
      if (!skipFlag) {
        guild['charArr'].push(estelDataArr[channel][nick]['name']);
      }
    } else {
      // キャラに登録されていないギルド
      // キャラが所属ギルド情報を違うギルドに再設定した可能性があるので
      // 登録済みならば削除しないとならない
      for (var i in guild['charArr']) {
        if (guild['charArr'][i] == estelDataArr[channel][nick]['name']) {
          // 登録済みなので削除するべし
          guild['charArr'].splice(i, 1);
          break;
        }
      }
      
    }
  }
  
  return true;
}


/*
 * ギルドデータを作成
 */
function createGuildData(channel, guildName)
{
  if (estelDataArr[channel] == undefined) {
    estelDataArr[channel] = [];
  }
  if (estelDataArr[channel]['guild'] == undefined) {
    estelDataArr[channel]['guild'] = [];
  }
  if (estelDataArr[channel]['guild'][guildName] == undefined) {
    estelDataArr[channel]['guild'][guildName] = [];
    estelDataArr[channel]['guild'][guildName]['charArr'] = [];
  }
}


/*
 * ギルド登録
 */
function estelGuild(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#guild(=|\s+)([^\s\/]+)\/(.+)$/i)) {
    return false;
  }
  
  var guildName = RegExp.$2;
  var guildSkillStr = RegExp.$3;
  
  // ギルドデータを作成
  createGuildData(channel, guildName);
  estelDataArr[channel]['guild'][guildName]['skillStr'] = guildSkillStr;
  
  // ギルド登録された際に、そのギルドに登録されているキャラクターを検索し
  // ギルドと結びつける
  for (var nick in estelDataArr[channel]) {
    connectCharToGuild(channel, nick);
  }
  
  var sendStr = 'ギルド "' + guildName + '" を登録しました。';
  estelSend(channel, sendStr);
  
  return true;
}


/*
 * ギルド一覧出力
 */
function estelGuildShow(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#gshow\s*$/i)) {
    return false;
  }
  
  if (estelDataArr[channel] == undefined
      || estelDataArr[channel]['guild'] == undefined) {
    return false;
  }
  
  var sendStr = '';
  for (var guildName in estelDataArr[channel]['guild']) {
    var guild = estelDataArr[channel]['guild'][guildName];
    if (guild['skillStr'] == undefined) {
      continue;
    }
    sendStr = guildName + ' : ' + guild['charArr'].join(', ');
    estelSend(channel, sendStr);
    sendStr = '    スキル : ' + guild['skillStr'];
    estelSend(channel, sendStr);
    if (guild['kago'] != undefined) {
      sendStr = '    加護 : ' + guild['kago'];
      estelSend(channel, sendStr);
    }
  }
  
  return true;
}


/*
 * ギルドデータ全部削除
 */
function estelGuildResetAll(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#gresetall\s*$/i)) {
    return false;
  }
  if (estelDataArr[channel] == undefined) {
    return false;
  }
  
  delete estelDataArr[channel]['guild'];
  
  var sendStr = 'ギルドデータをすべてリセットしました。';
  estelSend(channel, sendStr);
  
  return true;
}


/*
 * ==================================================
 * v1.1 GMサポート用エネミーデータ管理
 * ==================================================
 */


/*
 * estelDataArrにGMデータをセットする
 *
 * Perlと違ってJavaScriptは、利用前に配列初期化してやらないとエラーになるので
 * データセットするラッパー関数を用意
 */
function setEstelGMDataArr(channel, nick, key, val)
{
  if (estelDataArr[channel] == undefined) {
    estelDataArr[channel] = [];
  }
  if (estelDataArr[channel][GM_KEY] == undefined) {
    estelDataArr[channel][GM_KEY] = [];
  }
  if (estelDataArr[channel][GM_KEY]['nick'] != undefined
      && estelDataArr[channel][GM_KEY]['nick'] != nick) {
    return false;
  }
  
  estelDataArr[channel][GM_KEY][key] = val;
  
  return true;
}


/*
 * GMニックネームからチャンネルを引くための配列からデータ削除
 */
function deleteGMNickToChannel(channel)
{
  for (var gmNick in gmNickToChannelArr) {
    if (channel == gmNickToChannelArr[gmNick]) {
      delete gmNickToChannelArr[gmNick];
    }
  }
}


/*
 * GM登録
 */
function estelSetGM(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#setgm\s*$/i)) {
    return false;
  }
  
  // ニックネームチェック
  var sendStr = '';
  if (!prefix.nick.match(/GM/)) {
    sendStr = 'GM登録をするには、ニックネームの一部に"GM"という文字列を含めてください。';
    estelSend(channel, sendStr);
    return false;
  }
  
  // 他のチャンネルですでにGM登録されていないかどうかをチェック
  for (var gmNick in gmNickToChannelArr) {
    if (gmNick == prefix.nick) {
      sendStr = gmNick + 'はすでにGM登録されています。';
      estelSend(channel, sendStr);
      return false;
    }
  }
  
  setEstelGMDataArr(channel, prefix.nick, 'nick', prefix.nick);
  gmNickToChannelArr[prefix.nick] = channel;
  
  sendStr = prefix.nick + 'をこのチャンネルのGMとして設定しました。';
  estelSend(channel, sendStr);
  
  return true;
}


/*
 * GM削除
 */
function estelResetGM(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#resetgm\s*$/i)) {
    return false;
  }
  if (!prefix.nick.match(/GM/)) {
    return false;
  }
  if (estelDataArr[channel] == undefined) {
    return false;
  }
  
  delete estelDataArr[channel][GM_KEY];
  
  deleteGMNickToChannel(channel);
  
  var sendStr = 'このチャンネルのGM設定を削除しました。';
  estelSend(channel, sendStr);
  
  return true;
}


/*
 * 属性を表す文字の取得
 */
function getAttrStr(inAttrStr)
{
  switch (inAttrStr) {
  // 火
  case 'fire':  case 'f': case 'Fire':  case 'F': case '火':
    return '火';
    break;
  
  // 水
  case 'water': case 'w': case 'Water': case 'W': case '水':
    return '水';
    break;
  
  // 風
  case 'air':   case 'a': case 'Air':   case 'A': case '風':
    return '風';
    break;
  
  // 地
  case 'earth': case 'e': case 'Earth': case 'E': case '地':
    return '地';
    break;
  
  // 闇
  case 'dark':  case 'd': case 'Dark':  case 'D': case '闇':
    return '闇';
    break;
  
  // 光
  case 'light': case 'l': case 'Light': case 'L': case '光':
    return '光';
    break;
  }
  
  return '－';
}


/*
 * エネミーデータ登録
 */
function estelEnemySet(prefix, text)
{
  var channel = gmNickToChannelArr[prefix.nick];
  if (channel == undefined) {
    return false;
  }
  
  var regExp = new RegExp(
    "^#eset\\s+id=(\\d+)\\s+((" + 
      "((name|attr|shortname|uname)=[^\\s]+)|" + 
      "(cl|lv|fate|maxhp|hp|mhp|maxmp|mp|mmp|def|mdef)[+-]?=\\d+|" +
      "\\s" + 
    ")+)$", "i"
  );
  if (!text.match(regExp)) {
    return false;
  }
  
  var id = parseInt(RegExp.$1);
  var setArr = RegExp.$2.split(/\s+/);
  var sendStr;
  
  if (estelDataArr[channel][GM_KEY]['enemy'] == undefined) {
    estelDataArr[channel][GM_KEY]['enemy'] = [];
  }
  if (estelDataArr[channel][GM_KEY]['enemy'][id] == undefined) {
    estelDataArr[channel][GM_KEY]['enemy'][id] = [];
  }
  var enemyArr = estelDataArr[channel][GM_KEY]['enemy'][id];
  enemyArr['id'] = id;
  if (enemyArr['attr'] == undefined) {
    enemyArr['attr'] = '－';
  }
  
  for (var i in setArr) {
    // 値の設定(=)か、加減算(+=/-=)かを取得
    var vec = getVec(setArr[i]);
    
    // キーと値の取得
    var tmpArr = setArr[i].split(/[\+-]?=/);
    var key = tmpArr[0].toLowerCase();
    var val = tmpArr[1];
    
    // "lv"は"cl"に直す
    if (key == "lv") key = "cl";
    
    // "mhp"は"maxhp"に直す
    if (key == "mhp") key = "maxhp";
    
    // 最大HP"maxhp"の場合で、現在MPが設定されていない場合は
    // これも一緒に設定しておく（最初の一回のみ）
    if (key == "maxhp" && enemyArr['hp'] == undefined) {
      setDataAsInt(enemyArr, 'hp', val);
    }
    
    // "mmp"は"maxmp"に直す
    if (key == "mmp") key = "maxmp";
    
    // 最大MP"maxmp"の場合で、現在MPが設定されていない場合は
    // これも一緒に設定しておく（最初の一回のみ）
    if (key == "maxmp" && enemyArr['mp'] == undefined) {
      setDataAsInt(enemyArr, 'mp', val);
    }
    
    // フェイト"fate"の場合で、表記用のフェイト最大値が設定されていない場合には
    // これも一緒に設定しておく（最初の一回のみ）
    if (key == "fate" && enemyArr['maxfate'] == undefined) {
      setDataAsInt(enemyArr, 'maxfate', val);
    }
    
    // 属性部分
    if (key == 'attr') {
      val = getAttrStr(val);
      if (val == '－') {
        continue;
      }
    }
    
    if (vec != 0) {
      // 加減算の実行
      if (!calcEstelData(enemyArr, key, val, vec, prefix.nick)) {
        return false;
      }
    } else {
      // 値のセット
      setDataAsInt(enemyArr, key, val);
    }
  }
  
  outputData(prefix.nick, enemyArr, false);
  
  return true;
}


/*
 * エネミーデータ一覧出力
 */
function estelEnemyShow(prefix, text)
{
  if (!text.match(/^#eshow\s+id=?(\d+)\s*$/i)) {
    return false;
  }
  
  var id = parseInt(RegExp.$1);
  
  var channel = gmNickToChannelArr[prefix.nick];
  if (channel == undefined) {
    return false;
  }
  
  if (estelDataArr[channel][GM_KEY]['enemy'] == undefined
      || estelDataArr[channel][GM_KEY]['enemy'][id] == undefined) {
    return false;
  }
  
  outputData(prefix.nick, estelDataArr[channel][GM_KEY]['enemy'][id], false);
  
  return true;
}


/*
 * エネミーデータ一覧出力（全部）
 */
function estelEnemyShowAll(prefix, text)
{
  if (!text.match(/^#eshowall\s*$/i)) {
    return false;
  }
  
  var channel = gmNickToChannelArr[prefix.nick];
  if (channel == undefined) {
    return false;
  }
  
  if (estelDataArr[channel][GM_KEY]['enemy'] == undefined) {
    return false;
  }
  
  for (var id in estelDataArr[channel][GM_KEY]['enemy']) {
    outputData(prefix.nick, estelDataArr[channel][GM_KEY]['enemy'][id], false);
  }
  
  return true;
}


/*
 * エネミーデータリセット
 */
function estelEnemyReset(prefix, text)
{
  if (!text.match(/^#ereset\s*$/i)) {
    return false;
  }
  
  var channel = gmNickToChannelArr[prefix.nick];
  if (channel == undefined) {
    return false;
  }
  
  delete estelDataArr[channel][GM_KEY]['enemy'];
  
  var sendStr = 'エネミーデータをリセットしました。';
  estelSend(prefix.nick, sendStr);
  
  return true;
}


/*
 * エネミーにダメージ適用
 */
function estelEnemyDamage(prefix, text)
{
  var regExp = new RegExp(
    "^#e(m)?dam\\s+id=?(\\d+)\\s" + 
    "(\\s|(\\d+)|\\sp=?(\\d+)|\\s(nodef)|\\sattr=?([^\\s+]))+$", "i"
  );
  if (!text.match(regExp)) {
    return false;
  }
  
  var mdamFlag = RegExp.$1 != '';
  var id      = parseInt(RegExp.$2);
  var damage  = parseInt(RegExp.$4);
  var prevent = parseInt(RegExp.$5);
  var nodef   = RegExp.$6;
  var attrStr = RegExp.$7;
  var sendStr = '';
  
  var channel = gmNickToChannelArr[prefix.nick];
  if (channel == undefined
      || estelDataArr[channel][GM_KEY]['enemy'] == undefined
      || estelDataArr[channel][GM_KEY]['enemy'][id] == undefined
      || estelDataArr[channel][GM_KEY]['enemy'][id]['hp'] == undefined) {
    sendStr = 'エラー。現在HPの値が設定されていません。';
    estelSend(prefix.nick, sendStr);
    return false;
  }
  
  if (damage == '') {
    // ダメージの値がおかしい
    sendStr = 'エラー。ダメージの値が正しく設定されていません。';
    estelSend(prefix.nick, sendStr);
    return false;
  }
  
  var def = getDef(estelDataArr[channel][GM_KEY]['enemy'][id], mdamFlag);
  if (def < 0) {
    // 防御の値がおかしい
    var tmpStr = '物理';
    if (mdefFlag) {
      tmpStr = '魔法';
    }
    sendStr = 'エラー。' + tmpStr + '防御の値が正しく設定されていません。';
    estelSend(prefix.nick, sendStr);
    return false;
  }
  
  // 攻撃属性の取得
  attrStr = getAttrStr(attrStr);
  
  // 属性による防御点の変化
  var enemyAttr = estelDataArr[channel][GM_KEY]['enemy'][id]['attr'];
  if (attrStr != '－') {
    if (attrStr == enemyAttr) {
      // 同属性
      def *= 2;
    } else if (attrStr == againstAttrArr[enemyAttr]) {
      // 対抗属性
      def = 0;
    }
  }
  
  // 防御無視の適用
  if (nodef) {
    def = 0;
  }
  
  // HPダメージ値の決定
  if (isNaN(prevent)) prevent = 0;
  var hpDamage = damage - prevent - def;
  if (hpDamage < 0) {
    hpDamage = 0;
  }
  estelDataArr[channel][GM_KEY]['enemy'][id]['hp'] -= hpDamage;
  
  sendStr = getDamageSendStr(estelDataArr[channel][GM_KEY]['enemy'][id], 
                             def, mdamFlag, damage, prevent, 0, '', hpDamage,
                             attrStr, enemyAttr);
  
  estelSend(prefix.nick, sendStr);
  
  return true;
}


/*
 * ==================================================
 * v1.2 ランダムネームジェネレータ
 * ==================================================
 */
function estelGenerateRandomName(prefix, channel, text)
{
  if (!channel) {
    return false;
  }
  if (!text.match(/^#randomname\s+(男|女|man|woman|male|female|m|w|f)(\s+(\d+))?/)) {
    return false;
  }
  var isMale = true;
  var count  = 10;
  
  if (RegExp.$1) genderType = RegExp.$1;
  if (RegExp.$2) count = RegExp.$3;
  
  switch (genderType) {
  case '女':
  case 'woman':
  case 'female':
  case 'w':
  case 'f':
    isMale = false;
  }
  
  var req = new ActiveXObject("Microsoft.XMLHTTP");
  if (req) {
    req.onreadystatechange = function() {
      if (req.readyState == 4 && req.status == 200) {
        var namesListStr = req.responseText;
        var regExp = new RegExp("\n", "g");
        namesListStr = namesListStr.replace(regExp, "  ");
        
        var sendStr = 'こんな名前はどうでしょう？　' + namesListStr;
        estelSend(channel, sendStr);
      }
    }
    var param = 'gender=' + (isMale ? 'male' : 'female') + '&count=' + count;
    var url = 'http://www.g-chan.jp/trpg/rng/random_name_gen.php?' + param;
    req.open('get', url, true);
    req.send('');
  }
  
  return true;
}



/**
 * ここから簡単な単体テストコード
 */
function unitTest()
{
}


function unitTestLog(funcName, text)
{
}










/*
 * ここから、メルセンヌ・ツイスター部分
 * 
 * このjsファイルを提供してくれている、
 * http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/mt.html
 * によると、このjsファイルのライセンスは
 * 「mt.js は原典のライセンスに反しない限り自由に使用していただいて構いません」
 * とあり、原典はBSDライセンスなので、原典の著作権明記をした上で、
 * 再利用させていただいております。
 * 
 */

// mt.js 0.2.4 (2005-12-23)

/*

Mersenne Twister in JavaScript based on "mt19937ar.c"

 * JavaScript version by Magicant: Copyright (C) 2005 Magicant


 * Original C version by Makoto Matsumoto and Takuji Nishimura
   http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/mt.html

Copyright (C) 1997 - 2002, Makoto Matsumoto and Takuji Nishimura,
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

  3. The names of its contributors may not be used to endorse or promote 
     products derived from this software without specific prior written 
     permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/


// Methods whose name starts with "_" are private methods.
// Don't call them externally!


/**
 * Constructor: MersenneTwister([integer/Array<integer> seed])
 * initializes the object with the given seed.
 * The seed may be an integer or an array of integers.
 * If the seed is not given, the object will be initialized with the current
 * time: new Date().getTime().
 * See also: setSeed(seed).
 */
function MersenneTwister(seed) {
	if (arguments.length == 0)
		seed = new Date().getTime();
	
	this._mt = new Array(624);
	this.setSeed(seed);
}

/** multiplies two uint32 values and returns a uint32 result. */
MersenneTwister._mulUint32 = function(a, b) {
	var a1 = a >>> 16, a2 = a & 0xffff;
	var b1 = b >>> 16, b2 = b & 0xffff;
	return (((a1 * b2 + a2 * b1) << 16) + a2 * b2) >>> 0;
};

/** returns ceil(value) if value is finite number, otherwise 0. */
MersenneTwister._toNumber = function(x) {
	return (typeof x == "number" && !isNaN(x)) ? Math.ceil(x) : 0;
};

/**
 * Method: setSeed(integer/Array<integer> seed)
 * resets the seed. The seed may be an integer or an array of integers.
 * Elements in the seed array that are not numbers will be treated as 0.
 * Numbers that are not integers will be rounded down.
 * The integer(s) should be greater than or equal to 0 and less than 2^32.
 * This method is compatible with init_genrand and init_by_array function of
 * the original C version.
 */
MersenneTwister.prototype.setSeed = function(seed) {
	var mt = this._mt;
	if (typeof seed == "number") {
		mt[0] = seed >>> 0;
		for (var i = 1; i < mt.length; i++) {
			var x = mt[i-1] ^ (mt[i-1] >>> 30);
			mt[i] = MersenneTwister._mulUint32(1812433253, x) + i;
		}
		this._index = mt.length;
	} else if (seed instanceof Array) {
		var i = 1, j = 0;
		this.setSeed(19650218);
		for (var k = Math.max(mt.length, seed.length); k > 0; k--) {
			var x = mt[i-1] ^ (mt[i-1] >>> 30);
			x = MersenneTwister._mulUint32(x, 1664525);
			mt[i] = (mt[i] ^ x) + (seed[j] >>> 0) + j;
			if (++i >= mt.length) {
				mt[0] = mt[mt.length-1];
				i = 1;
			}
			if (++j >= seed.length) {
				j = 0;
			}
		}
		for (var k = mt.length - 1; k > 0; k--) {
			var x = mt[i-1] ^ (mt[i-1] >>> 30);
			x = MersenneTwister._mulUint32(x, 1566083941);
			mt[i] = (mt[i] ^ x) - i;
			if (++i >= mt.length) {
				mt[0] = mt[mt.length-1];
				i = 1;
			}
		}
		mt[0] = 0x80000000;
	} else {
		throw new TypeError("MersenneTwister: illegal seed.");
	}
};

/** returns the next random Uint32 value. */
MersenneTwister.prototype._nextInt = function() {
	var mt = this._mt, value;
	
	if (this._index >= mt.length) {
		var k = 0, N = mt.length, M = 397;
		do {
			value = (mt[k] & 0x80000000) | (mt[k+1] & 0x7fffffff);
			mt[k] = mt[k+M] ^ (value >>> 1) ^ ((value & 1) ? 0x9908b0df : 0);
		} while (++k < N-M);
		do {
			value = (mt[k] & 0x80000000) | (mt[k+1] & 0x7fffffff);
			mt[k] = mt[k+M-N] ^ (value >>> 1) ^ ((value & 1) ? 0x9908b0df : 0);
		} while (++k < N-1);
		value = (mt[N-1] & 0x80000000) | (mt[0] & 0x7fffffff);
		mt[N-1] = mt[M-1] ^ (value >>> 1) ^ ((value & 1) ? 0x9908b0df : 0);
		this._index = 0;
	}
	
	value = mt[this._index++];
	value ^=  value >>> 11;
	value ^= (value <<   7) & 0x9d2c5680;
	value ^= (value <<  15) & 0xefc60000;
	value ^=  value >>> 18;
	return value >>> 0;
};

/**
 * Method: nextInt([[number min,] number max])
 * returns a random integer that is greater than or equal to min and less than
 * max. The value of (max - min) must be positive number less or equal to 2^32.
 * If min is not given or not a number, this method uses 0 for min.
 * If neither of min and max is given or max is out of range, this method
 * uses 2^32 for max.
 * This method is compatible with genrand_int32 function of the original C
 * version for min=0 & max=2^32, but not with genrand_int31 function.
 */
MersenneTwister.prototype.nextInt = function() {
	var min, sup;
	switch (arguments.length) {
	case 0:
		return this._nextInt();
	case 1:
		min = 0;
		sup = MersenneTwister._toNumber(arguments[0]);
		break;
	default:
		min = MersenneTwister._toNumber(arguments[0]);
		sup = MersenneTwister._toNumber(arguments[1]) - min;
		break;
	}
	
	if (!(0 < sup && sup < 0x100000000))
		return this._nextInt() + min;
	if ((sup & (~sup + 1)) == sup)
		return ((sup - 1) & this._nextInt()) + min;
	
	var value;
	do {
		value = this._nextInt();
	} while (sup > 4294967296 - (value - (value %= sup)));
	return value + min;
};

/**
 * Method: next()
 * returns a random number that is greater than or equal to 0 and less than 1.
 * This method is compatible with genrand_res53 function of the original C
 * version.
 */
MersenneTwister.prototype.next = function() {
	var a = this._nextInt() >>> 5, b = this._nextInt() >>> 6;
	return (a * 0x4000000 + b) / 0x20000000000000; 
};


